#include "brasero-log.h"
#include <time.h>
#include <stdlib.h>
#include <glib.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <glib/gi18n-lib.h>

#define FILENAME_LONG_OVERSIZE 1024
#define LOG_MESSAGE_LONG_OVERSIZE 2048
#define MD5SUM_VALUE_SIZE 32

#define BRASERO_MD5_SUFFIX			".md5"
#define BRASERO_SHA1_SUFFIX			".sha1"
#define BRASERO_SHA256_SUFFIX		".sha256"

//int g_flag;
static int g_tp_connect_point;
static int g_log_fd;
static unsigned int locaLogID; 	/* brasero local log id */

static void
brasero_get_md5sum_value(char *filename, int len, char *value)
{
  char buf[FILENAME_LONG_OVERSIZE] = {0};
  char *output = NULL;

  g_snprintf (buf, FILENAME_LONG_OVERSIZE, "md5sum '%s'", filename);

  if (g_spawn_command_line_sync (buf, &output, NULL, NULL, NULL))
  {
    g_snprintf (value, len, "%s", output);
  }
}

/**
 * Get the last line of /var/log/nfs-burner/brasero.log.
 */
static char *
get_dpkg_log_file_last_line(void)
{
  const char *blog_file = "/var/log/nfs-burner/brasero.log";
  char tmp[FILENAME_LONG_OVERSIZE] = {0};

  if (access(blog_file, F_OK) != -1)
  {
    FILE *fp = fopen(blog_file, "r");

    while (!feof(fp))
    {
      fread(tmp, 1, sizeof(tmp), fp);
    }
    fclose(fp);
  }

  return g_strdup(tmp);
}

static void
remove_zero_l(char *dstr, char *sstr, gint len)
{
  int i = 0;
  char *tmp = g_malloc0(len * sizeof(gchar));

  strcpy(dstr, sstr);
  for (i = 0; dstr[i] == '0'; i++)
    ;

  strcpy(tmp, dstr + i);
  strcpy(dstr, tmp);
}

static void
init_local_log_id()
{
  char *last_line = get_dpkg_log_file_last_line();
  if (g_strstr_len(last_line, -1, ";;") != NULL)
  {
    char **strv_tmp1 = g_strsplit(last_line, ";;", -1);

    if (g_strstr_len(strv_tmp1[3], -1, "\"") != NULL)
    {
      char **strv_tmp2 = g_strsplit(strv_tmp1[3], "\"", -1);

      gint len = strlen(strv_tmp2[1]) + 1;

      char *id_str = g_malloc0(len * sizeof(gchar));

      remove_zero_l(id_str, strv_tmp2[1], len);
      if (id_str != NULL)
      {
        int id = atoi(id_str);
        locaLogID = (unsigned int)id;
      }
      g_strfreev(strv_tmp2);
    }
    g_strfreev(strv_tmp1);
  }
}

void open_log_file(void)
{
  int fd;

  init_local_log_id();

  if ((g_log_fd = open(LOCAL_LOG_FILE, O_RDWR | O_CREAT | O_APPEND, 0666)) == -1)
    perror("open /var/log/nfs-burner/brasero.log");

  if ((access(FILE_NAMES, F_OK)) == 0)
    remove(FILE_NAMES);
  if ((fd = open(FILE_NAMES, O_RDWR | O_CREAT | O_APPEND, 0666)) == -1)
    perror("open /var/log/nfs-burner/brasero.filename");
  else
    close(fd);
}

void write_local_log(const char *message)
{
  if (g_log_fd != -1)
  {
    char *local_log = calloc(1, strlen(message) + 4);
    sprintf(local_log, "%s\n", message);
    write(g_log_fd, local_log, strlen(local_log));
    fsync(g_log_fd);
    free(local_log);
  }
}

/*
void open_log_file(void) {
  openlog("brasero", LOG_PID|LOG_CONS, LOG_USER);
}

void write_local_log(const char *message) {
  syslog(LOG_INFO, "%s", message);
}
*/

int parse_xml_audit(const char *xml)
{
  int ret = -1;
  xmlDocPtr doc = NULL;
  xmlNodePtr node = NULL;
  xmlChar *value;

  if ((doc = xmlReadMemory(xml, strlen(xml), NULL, NULL, 0)) == NULL)
    return -1;

  if ((node = xmlDocGetRootElement(doc)) == NULL)
    return -1;

  if (xmlStrcmp(node->name, (const xmlChar *)"message"))
    return -1;

  node = node->xmlChildrenNode;
  while (node != NULL)
  {
    if (!(xmlStrcmp(node->name, (const xmlChar *)"body")))
    {
      node = node->xmlChildrenNode;
      while (node != NULL)
      {
        if (!(xmlStrcmp(node->name, (const xmlChar *)"TPRegisterstatus")))
        {
          value = xmlNodeGetContent(node);
          if (!(xmlStrcmp(value, (const xmlChar *)"1")))
            ret = 0;
          break;
        }
        node = node->next;
      }
      break;
    }
    node = node->next;
  }
  xmlFreeDoc(doc);

  return ret;
}

char _file_list[4096][FILENAME_LONG_OVERSIZE];
int _file_list_len = 1;

static int get_file_list(char *path, int depth)
{
  DIR *dir;
  struct dirent *file;
  struct stat st;

  if (!(dir = opendir(path)))
  {
    perror("Open dir error...");
    return -1;
  }

  gint length = strlen(path);

  while ((file = readdir(dir)) != NULL)
  {
    if (strcmp(file->d_name, ".") == 0 || strcmp(file->d_name, "..") == 0)
      continue;

    if (_file_list_len >= 4096)
      break;

    char tmp[FILENAME_LONG_OVERSIZE] = {0};
    gint pathname_size = g_snprintf(tmp, FILENAME_LONG_OVERSIZE, "%s/%s", path, file->d_name);
    strcpy(_file_list[_file_list_len++], tmp);

    if (file->d_type == 4)
    {
      if (length + strlen(file->d_name) >= FILENAME_LONG_OVERSIZE)
      {
        return -1;
      }
      // char child[FILENAME_LONG_OVERSIZE] = {0};
      // strcpy (child, path);
      // strcat (child, "/");
      gchar *child = g_strconcat(path, "/", file->d_name, NULL);
      get_file_list(child, ++depth);
    }
  }
  closedir(dir);
  return 0;
}

struct tm *current_time(void)
{
  struct tm *p_tm;
  time_t now;

  time(&now);
  p_tm = localtime(&now);

  return p_tm;
}

static char *get_file_suffix(const char *filename)
{
  char *dot = strrchr(filename, '.');
  if (!dot || dot == filename)
    return "其它";
  return dot + 1;
}

char *format_log_message(const char *filename, const char *dest_file,
                         const char *message, const char *display_name,
                         const char *dvd_or_cd, time_t used_time)
{
  char *ret_message;
  struct tm *p_tm;
  p_tm = current_time();
  char ret_message_list[LOG_MESSAGE_LONG_OVERSIZE] = {0};
  char md5sum_value[MD5SUM_VALUE_SIZE + 1] = {0};

  struct stat st;
  // float size;
  char filesize[20];
  memset(filesize, 0, 20);
  stat(filename, &st);
  if (S_ISREG(st.st_mode))
  {
    sprintf(filesize, "%.2fKB", ((float)st.st_size) / 1024);
  }
  char *usrname = getenv("USER");
  if (NULL == usrname)
  {
    return EXIT_FAILURE;
  }

  gchar *message_result = message;
  gchar message_operate[256];
  snprintf (message_operate, strlen(message) + 1, "%s", message);

  const gchar *msg_res_s = _("successfully");
  const gchar *msg_res_f = _("failed");
  const gchar *msg_res_u = _("unexpected");

  gchar *res = g_strrstr (message, msg_res_s);
  if (res)
  {
    if (!g_strcmp0(res, message))
    {
      snprintf (message_operate, strlen(message) - strlen(msg_res_s) + 1, "%s", message + strlen(msg_res_s));
    }
    else
    {
      snprintf (message_operate, strlen(message) - strlen(msg_res_s) + 1, "%s", message);
    }
    message_result = msg_res_s;
  }
  else
  {
    res = g_strrstr (message, msg_res_f);
    if (res)
    {
      if (!g_strcmp0(res, message))
      {
        snprintf (message_operate, strlen(message) - strlen(msg_res_f) + 1, "%s", message + strlen(msg_res_f));
      }
      else
      {
        snprintf (message_operate, strlen(message) - strlen(msg_res_f) + 1, "%s", message);
      }
      message_result = msg_res_f;
    }
    else
    {
      message_result = msg_res_u;
    }
  }

  char *suffix = get_file_suffix(filename);

  if (dest_file)
  {
    ret_message = g_malloc0((strlen(filename) + strlen(dest_file) + strlen(message) + 1024) * sizeof(gchar));
    if (S_ISREG(st.st_mode))
    {
      g_snprintf(ret_message, LOG_MESSAGE_LONG_OVERSIZE - 1, "时间=%d-%02d-%02d %02d:%02d:%02d, 用户名=%s, 刻录机类型=%s, 光盘属性=%s, 文件名称=%s (%s 文件, %s) , 操作=%s, 结果=%s",
                 p_tm->tm_year + 1900, p_tm->tm_mon + 1,
                 p_tm->tm_mday, p_tm->tm_hour,
                 p_tm->tm_min, p_tm->tm_sec,
                 usrname, display_name, dvd_or_cd, filename, suffix, filesize, message_operate, message_result);
    }
    else
    {
      g_snprintf(ret_message, LOG_MESSAGE_LONG_OVERSIZE - 1, "时间=%d-%02d-%02d %02d:%02d:%02d, 用户名=%s, 刻录机类型=%s, 光盘属性=%s, 文件名称=%s (%s 文件, %s) , 操作=%s, 结果=%s",
                 p_tm->tm_year + 1900, p_tm->tm_mon + 1,
                 p_tm->tm_mday, p_tm->tm_hour,
                 p_tm->tm_min, p_tm->tm_sec,
                 usrname, display_name, dvd_or_cd, filename, suffix, filesize, message_operate, message_result);
    }
  }
  else
  {
    if (S_ISREG(st.st_mode))
    {
      ret_message = g_malloc0((strlen(filename) + strlen(message) + 1024) * sizeof(gchar));
      g_snprintf(ret_message, LOG_MESSAGE_LONG_OVERSIZE - 1, "时间=%d-%02d-%02d %02d:%02d:%02d, 用户名=%s, 刻录机类型=%s, 光盘属性=%s, 文件名称=%s (%s 文件, %s) , 操作=%s, 结果=%s",
                 p_tm->tm_year + 1900, p_tm->tm_mon + 1,
                 p_tm->tm_mday, p_tm->tm_hour,
                 p_tm->tm_min, p_tm->tm_sec,
                 usrname, display_name, dvd_or_cd, filename, suffix, filesize, message_operate, message_result);
    }
    else
    {
      if (S_ISDIR(st.st_mode))
      {
        strcpy(_file_list[0], filename);
        int i;
        get_file_list(g_strdup(filename), 1);
        ret_message = g_malloc0(LOG_MESSAGE_LONG_OVERSIZE * _file_list_len * sizeof(gchar));
        for (i = 0; i < _file_list_len; i++)
        {
          if (stat(_file_list[i], &st) >= 0 && S_ISREG(st.st_mode))
          {
            sprintf(filesize, "%.2fKB", ((float)st.st_size) / 1024);
            g_snprintf(ret_message_list, LOG_MESSAGE_LONG_OVERSIZE - 1, "时间=%d-%02d-%02d %02d:%02d:%02d, 用户名=%s, 刻录机类型=%s, 光盘属性=%s, 文件名称=%s (%s 文件, %s) , 操作=%s, 结果=%s",
                      p_tm->tm_year + 1900, p_tm->tm_mon + 1,
                      p_tm->tm_mday, p_tm->tm_hour,
                      p_tm->tm_min, p_tm->tm_sec,
                      usrname, display_name, dvd_or_cd, _file_list[i], suffix, filesize, message_operate, message_result);
          }
          else
          {
            if (i == 0) {
              g_snprintf(ret_message_list, LOG_MESSAGE_LONG_OVERSIZE - 1, "时间=%d-%02d-%02d %02d:%02d:%02d, 用户名=%s, 刻录机类型=%s, 光盘属性=%s, 文件名称=%s (%s 文件, %s) , 操作=%s, 结果=%s",
                        p_tm->tm_year + 1900, p_tm->tm_mon + 1,
                        p_tm->tm_mday, p_tm->tm_hour,
                        p_tm->tm_min, p_tm->tm_sec,
                        usrname, display_name, dvd_or_cd, _file_list[i], suffix, filesize, message_operate, message_result);
            }
            else {
              g_snprintf(ret_message_list, LOG_MESSAGE_LONG_OVERSIZE - 1, "时间=%d-%02d-%02d %02d:%02d:%02d, 用户名=%s, 刻录机类型=%s, 光盘属性=%s, 文件名称=%s (%s 文件, %s) , 操作=%s, 结果=%s",
                        p_tm->tm_year + 1900, p_tm->tm_mon + 1,
                        p_tm->tm_mday, p_tm->tm_hour,
                        p_tm->tm_min, p_tm->tm_sec,
                        usrname, display_name, dvd_or_cd, _file_list[i], suffix, filesize, message_operate, message_result);
            }
          }
          if (i != _file_list_len - 1)
            strcat(ret_message_list, "\n");
          strncat(ret_message, ret_message_list, -1);
        }
        _file_list_len = 1;
      }
      else
      {
        ret_message = g_malloc0((strlen(filename) + strlen(message) + 1024) * sizeof(gchar));
        g_snprintf(ret_message, LOG_MESSAGE_LONG_OVERSIZE - 1, "时间=%d-%02d-%02d %02d:%02d:%02d, 用户名=%s, 刻录机类型=%s, 光盘属性=%s, 文件名称=%s (%s 文件, %s) , 操作=%s, 结果=%s",
                  p_tm->tm_year + 1900, p_tm->tm_mon + 1,
                  p_tm->tm_mday, p_tm->tm_hour,
                  p_tm->tm_min, p_tm->tm_sec,
                  usrname, display_name, dvd_or_cd, filename, suffix, filesize, message_operate, message_result);
      }
    }
  }
  return ret_message;
}

int tp_connect_audit_agent(void)
{
  int fd;
  struct sockaddr_un addr;
  fd = socket(AF_UNIX, SOCK_STREAM, 0);
  if (fd < 0)
  {
    close(fd);
    return -1;
  }
  bzero(&addr, sizeof(addr));
  addr.sun_family = AF_UNIX;
  strncpy(addr.sun_path, TP_CONNECT_POINT, sizeof(addr.sun_path));

  int ret = connect(fd, (struct sockaddr *)&addr, sizeof(addr));
  if (ret < 0)
  {
    close(fd);
    return -1;
  }

  return fd;
}

int tp_deal_register_respond(const char *tp_reg_respond)
{
  int ret;
  ret = parse_xml_audit(tp_reg_respond);

  return ret;
}

int tp_register_self(void)
{
  int ret;
  char buf[1024] = {0};
  char register_xml[1024] = {0};
  char *register_xml1 = "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
                        "<message>"
                        "<header>"
                        "<id/>"
                        "<from>brasero</from>"
                        "<to>tdsvc.pluginmgr</to>"
                        "<type>TPRegister</type>";
  char *register_xml2 = "</header>"
                        "<body>"
                        "<GUID>40E84C05-NFSC-4793-B84E-44255F76DDCC</GUID>"
                        "<TPName>brasero</TPName>"
                        "<TPDesc>create and copy CDs and DVDs</TPDesc>"
                        "</body>"
                        "</message>";

  struct tm *p_tm;

  p_tm = current_time();
  snprintf(register_xml, 1024, "%s<gentime>%d-%d-%d %d:%d:%d</gentime>%s",
           register_xml1, p_tm->tm_year + 1900,
           p_tm->tm_mon + 1, p_tm->tm_mday,
           p_tm->tm_hour, p_tm->tm_min,
           p_tm->tm_sec, register_xml2);

  ret = send(g_tp_connect_point, register_xml, strlen(register_xml), 0);
  if (ret <= 0)
  {
    printf("error send the register_xml");
    close(g_tp_connect_point);
    return -1;
  }

  ret = recv(g_tp_connect_point, buf, sizeof(buf), 0);
  if (ret <= 0)
  {
    printf("error recv the register respond");
    close(g_tp_connect_point);
    return -1;
  }

  ret = tp_deal_register_respond(buf);
  if (ret != 0)
  {
    printf("service is not send 1\n");
    close(g_tp_connect_point);
  }

  return ret;
}

int tp_setup_communication(void)
{
  int ret;
  g_tp_connect_point = tp_connect_audit_agent();
  if (g_tp_connect_point <= 0)
    return -1;

  ret = tp_register_self();

  return ret;
}

char *tp_format_audit_log(const char *message,
                          const char *display_name, const char *dvd_or_cd)
{
  char *ret_message = calloc(1, strlen(message) + FILENAME_LONG_OVERSIZE);
  char *ret_message1 = "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
                       "<message>"
                       "<header>"
                       "<id></id>"
                       "<from>brasero</from>"
                       "<to>tdsvr.log</to>"
                       "<type>TPAuditlog</type>";
  char *ret_message2 = "</header>"
                       "<body>"
                       "<Log type=\"TPAudit.安全刻录.3-6-1.中科方德\">";
  char *user = getenv("USER");
  if (user == NULL)
    return EXIT_FAILURE;
  char hostname[256] = {0};
  gethostname(hostname, FILENAME_LONG_OVERSIZE);
  char ret_message3[FILENAME_LONG_OVERSIZE] = {'\0'};
  sprintf(ret_message3, "<Sender type=\"服务\">create and copy CDs and DVDs</Sender>"
                        "<Trigger type=\"策略\" name=\"brasero\"/>"
                        "<Subject class=\"刻录软件\" name=\"brasero\">"
                        "<用户名>%s</用户名>"
                        "<计算机名称>%s</计算机名称>"
                        "<光盘类型>%s</光盘类型>"
                        "<刻录机类型>%s</刻录机类型>"
                        "</Subject>"
                        "<Subject class=\"User\" name=\"root\">"
                        "</Subject>"
                        "<Action>刻录光盘</Action>"
                        "<Object class=\"刻录\" name=\"文件\">",
          user, hostname, dvd_or_cd, display_name);
  char *ret_message4 = "</Object>"
                       "<Reason>未知</Reason>"
                       "<Level>2</Level>"
                       "<Handle>允许</Handle>"
                       "<Result>成功</Result>"
                       "</Log>"
                       "</body>"
                       "</message>";

  char time_message[512] = {0};
  struct tm *p_tm;
  p_tm = current_time();
  sprintf(time_message, "%d-%d-%d %d:%d:%d",
          p_tm->tm_year + 1900, p_tm->tm_mon + 1, p_tm->tm_mday,
          p_tm->tm_hour, p_tm->tm_min, p_tm->tm_sec);

  sprintf(ret_message, "%s<GenTime>%s</GenTime>"
                       "<SendTime>%s</SendTime>%s"
                       "<Time>%s</Time>%s"
                       "<TPAudit>\"%s\"</TPAudit>%s",
          ret_message1, time_message, time_message,
          ret_message2, time_message,
          ret_message3, message,
          ret_message4);

  return ret_message;
}

void tp_send_audit_log(const char *message,
                       const char *display_name, const char *dvd_or_cd)
{
  // if flag is not 0,then can not send message
  int flag;
  flag = tp_setup_communication();
  if (flag == 0)
  {
    char *tp_send_message;
    tp_send_message = tp_format_audit_log(message, display_name, dvd_or_cd);
    send(g_tp_connect_point, tp_send_message, strlen(tp_send_message), 0);
    free(tp_send_message);
    close(g_tp_connect_point);
  }
}

void brasero_init_log(void)
{
  open_log_file();
  /*	//if flag is not 0,then can not send message
  g_flag = tp_setup_communication();
  */
}

void brasero_send_log(char *filenames, char *dest_filenames,
                      const char *message, const char *display_name,
                      const char *dvd_or_cd, time_t used_time)
{
  if (filenames)
  {
    char filenames_arr[strlen(filenames) + 1];
    char *filename;

    // strcpy(filenames_arr, filenames);
    snprintf(filenames_arr, sizeof(filenames_arr), "%s", filenames);

    filename = strtok(filenames_arr, ";");
    while (filename)
    {
      if (g_str_has_suffix (filename, BRASERO_MD5_SUFFIX) || g_str_has_suffix (filename, BRASERO_SHA1_SUFFIX) || g_str_has_suffix (filename, BRASERO_SHA256_SUFFIX))
      {
        filename = strtok(NULL, ";");
        continue;
      }

      char *log_message;

      log_message = format_log_message(filename, dest_filenames,
                                       message, display_name, dvd_or_cd, used_time);
      write_local_log(log_message);
      //#if 0
      tp_send_audit_log(log_message, display_name, dvd_or_cd);
      //#endif
      free(log_message);

      filename = strtok(NULL, ";");
    }
  }
  else
  {
    char *log_message;

    log_message = format_log_message(dest_filenames, NULL,
                                     message, display_name, dvd_or_cd, used_time);
    write_local_log(log_message);
    //#if 0
    tp_send_audit_log(log_message, display_name, dvd_or_cd);
    //#endif
    free(log_message);
  }

  char buf[FILENAME_LONG_OVERSIZE];
  int fd;
  FILE *fp;
  fp = fopen(FILE_NAMES, "r");
  if (fp)
  {
    while ((fgets(buf, FILENAME_LONG_OVERSIZE, fp)) != NULL)
    {
      int len;
      char *log_message;

      len = strlen(buf);
      buf[len - 1] = '\0'; // delete \n
      log_message = format_log_message(buf, dest_filenames,
                                       message, display_name, dvd_or_cd, used_time);
      write_local_log(log_message);
      //#if 0
      tp_send_audit_log(log_message, display_name, dvd_or_cd);
      //#endif
      free(log_message);
    }
    fclose(fp);
  }

  if ((access(FILE_NAMES, F_OK)) == 0)
    remove(FILE_NAMES);
  if ((fd = open(FILE_NAMES, O_RDWR | O_CREAT | O_APPEND, 0666)) == -1)
    perror("open /var/log/nfs-burner/brasero.filename");
  else
    close(fd);
}
